import 'dart:async';
import 'dart:ffi';
import 'dart:core';

import 'package:unit_rc/UnitRC.dart';

enum ProxyRCType {
  none,
  strong,
  weak,
}

mixin DartBaseProxyClass on DartBaseClass {
  DartBaseClass? strongBaseRC() {
    return _native_strongBaseRC(toCpp()) as DartBaseClass?;
  }
  static late final _native_strongBaseRC = unitRCLib.lookupFunction<
    Handle Function(Pointer<Void>),
    Object? Function(Pointer<Void>)
  >("UnitRC_strongBaseRC");

  DartBaseClass? weakBaseRC() {
    return _native_weakBaseRC(toCpp()) as DartBaseClass?;
  }
  static late final _native_weakBaseRC = unitRCLib.lookupFunction<
    Handle Function(Pointer<Void>),
    Object? Function(Pointer<Void>)
  >("UnitRC_weakBaseRC");

  bool get isProxy {
    return true;
  }
}

class WeakPtr<T extends DartBaseClass> {
  WeakReference<T>? weakPtr;
  T? strongPtr;

  WeakPtr({T? ptr}) {
    if (ptr is DartBaseProxyClass) {
      this.strongPtr = ptr;
    } else {
      if (ptr != null) {
        this.weakPtr = WeakReference(ptr);
      }
    }
  }

  T? lock() {
    return (strongPtr??(weakPtr?.target))?.strongRC();
  }
}

class DartBaseClass {
  final zone = Zone.current;

  Pointer<Void> strongPointer = nullptr;
  Pointer<Void> weakPointer = nullptr;

  Pointer<Void> toCpp({ProxyRCType rcType = ProxyRCType.strong}) {
    final isProxy = this.isProxy;
    if (strongPointer != nullptr && (isProxy || rcType == ProxyRCType.strong)) {
      return strongPointer;
    }
    if (weakPointer != nullptr && (isProxy || rcType == ProxyRCType.weak)) {
      return weakPointer;
    }
    return nullptr;
  }

  bool get isProxy {
    return false;
  }

  void notifyFieldUpdated(int index) {
    _native_notifyFieldUpdated(toCpp(), index);
  }
  static late final _native_notifyFieldUpdated = unitRCLib.lookupFunction<
      Void Function(Pointer<Void>, Int32),
      void Function(
          Pointer<Void>, int)>("UnitRC_DartBaseClass_notifyFieldUpdated");
}

extension UnitRCWeakEx<T extends DartBaseClass> on T {
  T? strongRC() {
    if (this is DartBaseProxyClass) {
      return (this as DartBaseProxyClass).strongBaseRC() as T?;
    }
    return this;
  }

  WeakPtr<T> weakRC() {
    if (this is DartBaseProxyClass) {
      return WeakPtr(ptr: (this as DartBaseProxyClass).weakBaseRC() as T?);
    }
    return WeakPtr<T>(ptr: this);
  }
}